/********************************************************************
 File:  EPImporter.m
 
 Created: 5/6/10
 Modified: 7/15/10
 
 Author: Yunfeng Zhang
 Cognitive Modeling and Eye Tracking Lab
 CIS Department
 University of Oregon
 
 Funded by the Office of Naval Research & National Science Foundation.
 Primary Investigator: Anthony Hornof.
 
 Copyright (c) 2010 by the University of Oregon.
 ALL RIGHTS RESERVED.
 
 Permission to use, copy, and distribute this software in
 its entirety for non-commercial purposes and without fee,
 is hereby granted, provided that the above copyright notice
 and this permission notice appear in all copies and their
 documentation.
 
 Software developers, consultants, or anyone else who wishes
 to use all or part of the software or its documentation for
 commercial purposes should contact the Technology Transfer
 Office at the University of Oregon to arrange a commercial
 license agreement.
 
 This software is provided "as is" without expressed or
 implied warranty of any kind.
 ********************************************************************/


#import "EPImporter.h"
#import "RegexKitLite.h"

@implementation EPImporter

@synthesize moc;


- (id)initWithMOC:(NSManagedObjectContext *)anMOC
{	
	if (self = [super init]) {
		moc = anMOC;
    }
    return self;
}

- (void)import:(NSURL *)rawDataFileURL
{
	printf("Starting to import file %s...\n", [[rawDataFileURL path] UTF8String]);
	NSError *error;
	
	// Read settings.
	NSString *settingPath = [[[rawDataFileURL path] stringByDeletingLastPathComponent] 
													stringByAppendingPathComponent:@"settings.txt"];
	
	// Windows uses ANSCII
	NSString *fileContent = [NSString stringWithContentsOfURL:[NSURL fileURLWithPath:settingPath] 
												  encoding:NSUTF8StringEncoding 
												 	 error:&error];
	NSArray *lines = [fileContent componentsSeparatedByString:@"\n"];
	
	VFSession *session = [NSEntityDescription insertNewObjectForEntityForName:@"Session" inManagedObjectContext:moc];
	
	for (NSString *eachline in lines) {
		NSArray *fields = [eachline componentsSeparatedByString:@": "];
		if ([fields count] == 2) {
			NSString * settingName = [fields objectAtIndex:0];
			NSString * settingValue = [fields objectAtIndex:1];
			if ([settingName isEqualToString:@"EyeToScreenDisctance"]) {
				session.distanceToScreen = [NSNumber numberWithInt:[settingValue intValue]]; // in mm.
			} else if ([settingName isEqualToString:@"ScreenResolution"]) {
				NSArray *resolutions = [settingValue componentsSeparatedByString:@", "];
				session.screenResolution = NSMakeSize([[resolutions objectAtIndex:0] floatValue], 
													  [[resolutions objectAtIndex:1] floatValue]);
			} else if ([settingName isEqualToString:@"ScreenDimension"]) {
				NSArray *dimensions = [settingValue componentsSeparatedByString:@", "];
				session.screenDimension = NSMakeSize([[dimensions objectAtIndex:0] floatValue], 
													  [[dimensions objectAtIndex:1] floatValue]);
			} if ([settingName isEqualToString:@"GazeSampleRate"]) {
				session.gazeSampleRate = [NSNumber numberWithInt:[settingValue intValue]];
			}
		}
	}
	
	// Read experiment info.
	NSString *infoPath = [[[rawDataFileURL path] stringByDeletingPathExtension] 
							 stringByAppendingPathExtension:@"txt"];
	// EPrime uses UTF-16
	fileContent = [NSString stringWithContentsOfURL:[NSURL fileURLWithPath:infoPath] 
													encoding:NSUTF16LittleEndianStringEncoding 
													error:&error];
	NSString *stars = @"\\*\\*\\*";
	
	NSString *pattern = [NSString stringWithFormat:@"(?s)%@ Header Start %@\\r\\n(.+)%@ Header End %@\r\n(.+)Level: 1", stars, stars, stars, stars];
	NSString *headString, *bodyString;
	headString = [fileContent stringByMatching:pattern capture:1];
	bodyString = [fileContent stringByMatching:pattern capture:2];
	
	// Parse date.
	pattern = @"SessionDate: (\\d\\d-\\d\\d-\\d{4})";
	NSString *dateStr = [headString stringByMatching:pattern capture:1];
	pattern = @"SessionTime: (\\d\\d:\\d\\d:\\d\\d)";
	NSString *timeStr = [headString stringByMatching:pattern capture:1];
	
	NSDateFormatter *inputFormatter = [[NSDateFormatter alloc] init];
	[inputFormatter setDateFormat:@"MM-dd-yyyy HH:mm:ss"];
	session.date = [inputFormatter dateFromString:[NSString stringWithFormat:@"%@ %@", dateStr, timeStr]];
	
	// Parse experiment name;
	pattern = @"Experiment: (.+)";
	session.experiment = [headString stringByMatching:pattern capture:1];
	
	// Parse subject ID;
	pattern = @"Subject: (\\d+)";
	session.subjectID = [headString stringByMatching:pattern capture:1];
	
	// Parse session ID;
	pattern = @"Session: (\\d+)";
	session.sessionID = [headString stringByMatching:pattern capture:1];
		
	// Parse body
	pattern = [NSString stringWithFormat:@"(?s)Level: 2\\s+%@ LogFrame Start %@\\r\\n(.+?)\\t%@ LogFrame End %@", stars, stars, stars, stars];
	NSArray *trialStrs = [bodyString componentsMatchedByRegex:pattern];
	
	NSMutableArray *allTrials = [NSMutableArray array];
	
	for (NSString *trialStr in trialStrs) {
		pattern = @"Running: (\\w+)";
		if (![[trialStr stringByMatching:pattern capture:1] isEqualToString:@"PracticeTrialList"]) {
		
			VFProcedure *trial = [NSEntityDescription insertNewObjectForEntityForName:@"Procedure" inManagedObjectContext:moc];
			
			pattern = @"Stimulus.OnsetTime: (\\d+)";
			trial.startTime = [NSNumber numberWithInt:[[trialStr stringByMatching:pattern capture:1] intValue]];
			
			pattern = @"Stimulus.OffsetTime: (\\d+)";
			NSString * matchedTime = [trialStr stringByMatching:pattern capture:1];
			if (matchedTime == nil) {
				pattern = @"Stimulus.RTTime: (\\d+)";
				trial.endTime = [NSNumber numberWithInt:[[trialStr stringByMatching:pattern capture:1] intValue]];	
			} else {
				trial.endTime = [NSNumber numberWithInt:[matchedTime intValue]];
			}
			
			pattern = @"TrialList: (\\d+)";
			trial.ID = [trialStr stringByMatching:pattern capture:1];
			
			[allTrials addObject:trial];
		}
	}
	
	session.duration = [[allTrials lastObject] endTime];
	
	// Read screenshots.
	NSString *screenshotsPath = [[[rawDataFileURL path] stringByDeletingPathExtension] 
									stringByAppendingPathExtension:@"screenshots"];
	
	fileContent = [NSString stringWithContentsOfURL:[NSURL fileURLWithPath:screenshotsPath] 
										   encoding:NSUTF8StringEncoding 
											  error:&error];
	lines = [fileContent componentsSeparatedByString:@"\n"];
	
	for (NSString *eachline in lines) {
		NSArray *fields = [eachline componentsSeparatedByString:@": "];
		if ([fields count] == 2) {
			NSString * trialID = [fields objectAtIndex:0];
			NSString * trialScreenshotPath = [fields objectAtIndex:1];
			
			for (VFProcedure *eachTrial in allTrials) {
				if ([eachTrial.ID isEqualToString:trialID]) {
					VFVisualStimulus *screenshot = [NSEntityDescription insertNewObjectForEntityForName:@"VisualStimulus" 
																				 inManagedObjectContext:moc];
					screenshot.startTime = eachTrial.startTime;
					screenshot.endTime = eachTrial.endTime;					
					screenshot.ID = trialID;
					
					VFVisualStimulusFrame *frame = [NSEntityDescription insertNewObjectForEntityForName:@"VisualStimulusFrame" 
																				 inManagedObjectContext:moc];
					frame.location = NSMakePoint(0, 0);
					frame.startTime = screenshot.startTime;
					frame.endTime = screenshot.endTime;
					
					[screenshot addFramesObject:frame];
					
					
					VFVisualStimulusTemplate *template = [NSEntityDescription insertNewObjectForEntityForName:@"VisualStimulusTemplate" 
																					   inManagedObjectContext:moc];
					
					template.imageFilePath = trialScreenshotPath;
					template.zorder = [NSNumber numberWithInt:0];
					template.category = @"background";
				
					screenshot.template = template;
					
					[allTrials removeObject:eachTrial];
					break;
				}
			}
		}
	}
	
	
	// Read eye movements.
	fileContent = [NSString stringWithContentsOfURL:rawDataFileURL
										   encoding:NSASCIIStringEncoding 
											  error:&error];
	lines = [fileContent componentsSeparatedByString:@"\r\n"];
	
	NSString *header = [lines objectAtIndex:0];
	int timeIndex, xIndex, yIndex, i = 0;
	
	for (NSString *field in [header componentsSeparatedByString:@"\t"]) {
		if ([field isEqualToString:@"RTTime"]) {
			timeIndex = i;
		} else if ([field isEqualToString:@"CursorX"]) {
			xIndex = i;
		} else if ([field isEqualToString:@"CursorY"]) {
			yIndex = i;
		}
		++i;
	}
	
	for (NSString *eachline in [lines objectsAtIndexes:
								[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(1, [lines count] - 1)]]) {
		NSArray *fields = [eachline componentsSeparatedByString:@"\t"];
		
		if ([fields count] == i) {
			VFGazeSample *gaze = [NSEntityDescription insertNewObjectForEntityForName:@"GazeSample" 
															   inManagedObjectContext:moc];
			
			gaze.location = NSMakePoint([[fields objectAtIndex:xIndex] floatValue], [[fields objectAtIndex:yIndex] floatValue]);
			gaze.valid = [NSNumber numberWithBool:(gaze.location.x == -1 ? NO : YES)];
			gaze.time = [NSNumber numberWithInt:[[fields objectAtIndex:timeIndex] intValue]];
		}
	}
	
	[self saveData];
	
	NSLog(@"Import file %@ succeeded.", [rawDataFileURL path]);
	
	NSLog(@"Start to detect fixations.");
	
	VFFetchHelper *fetchHelper = [[VFFetchHelper alloc]	initWithMOC:moc];
	NSArray *gazeArray = [fetchHelper fetchAllObjectsForName:@"GazeSample"];
	[VFDTFixationAlg detectFixation:gazeArray withDispersionThreshold:0.9 andMinFixationDuration:100];
	
	[self saveData];
	
	NSLog(@"Detecting fixations succeeded.");
	
	printf("done.\n");
	
}

- (void)saveData
{
	NSError *error = nil;
	if (![moc save:&error])
	{
		NSLog(@"Data Import Failure\n%@",
			  ([error localizedDescription] != nil) ?
			  [error localizedDescription] : @"Unknown Error");
		NSArray *detailedErrors = [[error userInfo] objectForKey:NSDetailedErrorsKey];
		NSLog(@"%u validation errors have occurred", [detailedErrors count]);
		for (NSError *detailError in detailedErrors) {
			NSLog(@"%@\n%@", [detailError localizedDescription], [[detailError userInfo] objectForKey:NSValidationObjectErrorKey]);
		}
	}
}

@end
